@@ -1,9 +1,9 @@
 // -*- c++ -*-
 //
-// $Id: PctestIpv4Raw.h,v 1.4 2000/01/19 17:14:42 bmah Exp $
+// $Id: PctestIpv4Raw.h,v 1.7 2001/01/02 18:46:04 bmah Exp $
 //
 // PctestIpv4Raw.h
-// Bruce A. Mah <bmah@ca.sandia.gov>
+// Bruce A. Mah <bmah@acm.org>
 //
 // This work was first produced by an employee of Sandia National
 // Laboratories under a contract with the U.S. Department of Energy.
@@ -14,7 +14,7 @@
 // others. This software is provided as-is. SANDIA DISCLAIMS ANY
 // WARRANTY OF ANY KIND, EXPRESS OR IMPLIED.
 //
-// Header class for IPv4 tests using raw sockets
+// Header class for IPv4 tests using raw sockets to send UDP probes.
 //
 
 #ifndef PCTESTIPV4RAW_H
@@ -37,29 +37,17 @@
 #endif /* HAVE_STRINGS_H */
 
 #include "pc.h"
-#include "PctestIpv4.h"
+#include "PctestIpv4Udp.h"
 
-class PctestIpv4Raw : public PctestIpv4 {
+class PctestIpv4Raw : public PctestIpv4Udp {
 
   public:
 
     PctestIpv4Raw() { 
-//	socketOut = 0;
-//	socketIn = 0;
-//	destPort = 32768; 
     };
     PctestIpv4Raw(int p) { 
-//	socketOut = 0;
-//	socketIn = 0;
-//	destPort = p; 
     };
     virtual ~PctestIpv4Raw() { 
-//	if (socketOut > 0) {
-//	    close(socketOut);
-//	}
-//	if (socketIn > 0) {
-//	    close(socketIn);
-//	}
     };
 
     virtual int GetSocketOut();
