@@ -1,9 +1,9 @@
 // -*- c++ -*-
 //
-// $Id: PctestIpv4Udp.h,v 1.4 2000/01/19 17:14:42 bmah Exp $
+// $Id: PctestIpv4Udp.h,v 1.8 2001/01/02 18:46:08 bmah Exp $
 //
 // PctestIpv4Udp.h
-// Bruce A. Mah <bmah@ca.sandia.gov>
+// Bruce A. Mah <bmah@acm.org>
 //
 // This work was first produced by an employee of Sandia National
 // Laboratories under a contract with the U.S. Department of Energy.
@@ -44,26 +44,16 @@ class PctestIpv4Udp : public PctestIpv4 {
   public:
 
     PctestIpv4Udp() { 
-//	socketOut = 0;
-//	socketIn = 0;
-//	destPort = 32768; 
     };
     PctestIpv4Udp(int p) { 
-//	socketOut = 0;
-//	socketIn = 0;
-//	destPort = p; 
     };
     virtual ~PctestIpv4Udp() { 
-//	if (socketOut > 0) {
-//	    close(socketOut);
-//	}
-//	if (socketIn > 0) {
-//	    close(socketIn);
-//	}
     };
 
     virtual int GetSocketOut();
     virtual int Test(TestRecord &tr);
+    virtual int GetMinSize();
+    virtual PctestActionType GetAction(int icmpType, int icmpCode);
    
   protected:
 
