@@ -1,9 +1,9 @@
-static char rcsid[] = "$Id: PctestIpv6Udp.cc,v 1.13 2000/05/05 22:43:27 bmah Exp $";
+static char rcsid[] = "$Id: PctestIpv6Udp.cc,v 1.24 2001/01/22 17:36:00 bmah Exp $";
 //
-// $Id: PctestIpv6Udp.cc,v 1.13 2000/05/05 22:43:27 bmah Exp $
+// $Id: PctestIpv6Udp.cc,v 1.24 2001/01/22 17:36:00 bmah Exp $
 //
 // PctestIpv6Udp.cc
-// Bruce A. Mah <bmah@ca.sandia.gov>
+// Bruce A. Mah <bmah@acm.org>
 //
 // This work was first produced by an employee of Sandia National
 // Laboratories under a contract with the U.S. Department of Energy.
@@ -49,18 +49,16 @@ static char rcsid[] = "$Id: PctestIpv6Udp.cc,v 1.13 2000/05/05 22:43:27 bmah Exp
 //
 // Output:  In return value, returns socket number.
 //
-// Get output socket of an appropriate type.  We'll make this virtual 
-// shortly.
+// Get output socket of an appropriate type, store in socketOut.
 //
 int PctestIpv6Udp::GetSocketOut() {
 
-    int sock;
     int rc;
     
-    sock = socket(AF_INET6, SOCK_DGRAM, 0);
-    if (sock < 0) {
+    socketOut = socket(AF_INET6, SOCK_DGRAM, 0);
+    if (socketOut < 0) {
 	perror("socket");
-	return sock;
+	return socketOut;
     }
     
 #ifdef linux
@@ -68,7 +66,7 @@ int PctestIpv6Udp::GetSocketOut() {
     // getting ICMP errors when we send packets out.
     int bsdcompatOption;
     bsdcompatOption = 1;
-    rc = setsockopt(sock, SOL_SOCKET, SO_BSDCOMPAT, &bsdcompatOption,
+    rc = setsockopt(socketOut, SOL_SOCKET, SO_BSDCOMPAT, &bsdcompatOption,
 		    sizeof(bsdcompatOption));
     if (rc < 0) {
 	perror("setsockopt(SO_BSDCOMPAT)");
@@ -76,7 +74,27 @@ int PctestIpv6Udp::GetSocketOut() {
     }
 #endif /* linux */
 
-    return sock;
+    // Make up a socket address structure for the source address
+    // and attempt to bind the output socket to it.
+    struct sockaddr_in6 originSocketAddress;
+    memset((void *) &originSocketAddress, 0, sizeof(struct sockaddr_in6));
+    // See comments in PctestIpv6::SetTargetName() about the overloading
+    // of HAVE_SOCKADDR_SA_LEN.
+#ifdef HAVE_SOCKADDR_SA_LEN
+    originSocketAddress.sin6_len = sizeof(struct sockaddr_in6);
+#endif /* HAVE_SOCKADDR_SA_LEN */
+    originSocketAddress.sin6_family = AF_INET6;
+    originSocketAddress.sin6_port = htons(0);
+    memcpy(&(originSocketAddress.sin6_addr), &originAddress, sizeof(in6_addr));
+
+    rc = bind(socketOut, (struct sockaddr *) &originSocketAddress, 
+	      sizeof(originSocketAddress));
+    if (rc < 0) { 
+	perror("bind()");
+	return rc;
+    }
+
+    return socketOut;
 }
 
 //
@@ -164,9 +182,9 @@ int PctestIpv6Udp::Test(TestRecord &tr)
 
 	// Read response from socket
 	if (rc == 1) {
-	    IF_DEBUG(2, fprintf(stderr, "icmp packet received\n"));
+	    IF_DEBUG(2, fprintf(stderr, "Response packet received\n"));
 
-	    const int icmpPacketSize = 1500;
+	    const int icmpPacketSize = 1500 + tr.size;
 	    char icmpPacket[icmpPacketSize];
 	    struct msghdr msg;	// msghdr is for recvmsg
 	    struct iovec iov[1];
@@ -205,6 +223,7 @@ int PctestIpv6Udp::Test(TestRecord &tr)
 		perror("read");
 		return rc;
 	    }
+	    tr.replsize = rc;
 
 	    // Now parse the packet, doing a little error checking along
 	    // the way.  By the end, we'll have ipHeader and icmpHeader
@@ -215,71 +234,80 @@ int PctestIpv6Udp::Test(TestRecord &tr)
 	    icmp6_hdr *icmpHeader;
 	    udphdr *udpHeader;
 
+	    if (tr.replsize - (0) < sizeof(icmp6_hdr)) {
+		IF_DEBUG(3, fprintf(stderr, "Received incomplete ICMPv6 packet of %d bytes\n", tr.replsize));
+		continue;
+	    }
+
 	    // Find the ICMPv6 header
 	    icmpHeader = (icmp6_hdr *) icmpPacket;
-	    IF_DEBUG(3, fprintf(stderr, "icmp type = %d, code = %d\n", 
+	    IF_DEBUG(3, fprintf(stderr, "ICMPv6 type = %d, code = %d\n", 
 				icmpHeader->icmp6_type, icmpHeader->icmp6_code));
 
-	    // Check ICMP type.  Most types (such as echo request/reply,
-	    // router adverts, etc.) we ignore.
+	    // Check ICMPv6 type.  Note that we already set up ICMPv6
+	    // message filtering in PctestIpv6::GetSocketIn(), which
+	    // should catch most messages we don't want.  This check
+	    // handles the case of messages that pchar in general is
+	    // interested in, but this particular test is not.  We may
+	    // want to rethink this in the future, perhaps by making
+	    // ICMPv6 filtering dependent on the test protocol in use..
 	    if ((icmpHeader->icmp6_type != ICMP6_TIME_EXCEEDED) && 
 		(icmpHeader->icmp6_type != ICMP6_DST_UNREACH)) {
-		IF_DEBUG(3, fprintf(stderr, "ignoring icmp packet\n"));
+		IF_DEBUG(3, fprintf(stderr, "Ignoring ICMPv6 packet\n"));
 		goto donepacket;
 	    }
 	    
-	    // Check for a valid (to us) IP header within the packet.
-	    // For "time exceeded" or "destination unreachable", this
-	    // header will be 8 bytes past the ICMP header.
-	    ipHeader2 = (ip6_hdr *) ((char *) icmpHeader + sizeof(icmp6_hdr));
-
-	    // Check to be sure that we have enough of the packet to hold
-	    // a valid IP header? XXX
-
-	    // Additional checking here...must be UDP
-	    if (ipHeader2->ip6_nxt != IPPROTO_UDP) {
-		IF_DEBUG(3, fprintf(stderr, "ignoring icmp packet for non-udp\n"));
+	    if (tr.replsize - (0 + sizeof(icmp6_hdr)) < 
+		sizeof(ip6_hdr)) {
+		IF_DEBUG(3, fprintf(stderr, "Received incomplete inner IPv6 packet, %d bytes total\n", tr.replsize));
 		goto donepacket;
 	    }
 
-	    // Align UDP header template, check port numbers and
-	    // payload length for us. XXX
+	    // Check for a valid (to us) IP header within the packet.
+	    // For "time exceeded" or "destination unreachable", this
+	    // header will be just past the ICMPv6 header.
+	    ipHeader2 = (ip6_hdr *) ((char *) icmpHeader + sizeof(icmp6_hdr));
+
+	    // Note:  We can look for a UDP header immediately following
+	    // the inner IPv6 header because we know (or at least we
+	    // think we know) that the original packet went out with no
+	    // extension headers.  If that's not true, this test will fail.
+	    // If we happen to be a situation where we need to do this,
+	    // then we have to insert a parser to skip over the extension
+	    // headers right here.
+	    if (ipHeader2->ip6_nxt != IPPROTO_UDP) {
+		IF_DEBUG(3, fprintf(stderr, "Ignoring ICMPv6 packet for non-UDP\n"));
+		goto donepacket;
+	    }
+
+	    // Check to see if we got enough for a complete UDP header.
+	    // RFC 2463 says that we should get back as much of the
+	    // packet that will fit in an MTU.
+	    if (tr.replsize - (0 + sizeof(icmp6_hdr) + sizeof(ip6_hdr)) <
+		sizeof(udphdr)) {
+		IF_DEBUG(3, fprintf(stderr, "Received incomplete inner UDP packet, %d bytes total\n", tr.replsize));
+		goto donepacket;
+	    }
+
+	    // Align UDP header template.
 	    udpHeader = (udphdr *) (((char *) ipHeader2) + sizeof(ip6_hdr));
 
 	    // Check destination UDP port number (we don't know the
 	    // source) and UDP (header+payload) length
 	    if ((udpHeader->uh_dport != targetSocketAddress.sin6_port) || 
 		(ntohs(udpHeader->uh_ulen) != udpPayloadSize + sizeof(udphdr))) {
-		IF_DEBUG(3, fprintf(stderr, "ignoring icmp packet for unknown udp packet\n"));
+		IF_DEBUG(3, fprintf(stderr, "Ignoring ICMP packet for unknown UDP packet\n"));
 		goto donepacket;
 	    }
 
 	    // Fill in return fields
-	    tr.targetAddress = new char[sizeof(in6_addr)];
-	    memcpy(tr.targetAddress, &targetAddress, sizeof(in6_addr));
-	    tr.targetAddressLength = sizeof(in6_addr);
-
-	    // See if there's any ancillary data...if so it should
-	    // be the receiving address for the ICMPv6 packet.  Hopefully
-	    // that's the same as the source address for the original
-	    // UDP packet we sent.
-	    tr.icmpDestAddress = new char[sizeof(in6_addr)];
-	    memset(tr.icmpDestAddress, 0, sizeof(in6_addr));
-	    cmsgptr = CMSG_FIRSTHDR(&msg);
-	    if (cmsgptr) {
-		if ((cmsgptr->cmsg_level == IPPROTO_IPV6) && 
-		    (cmsgptr->cmsg_type == IPV6_PKTINFO)) {
-		    memcpy(tr.icmpDestAddress, CMSG_DATA(cmsgptr), sizeof(in6_addr));
-		}
-	    }
-	    tr.icmpDestAddressLength = sizeof(in6_addr);
-
 	    tr.icmpSourceAddress = new char[sizeof(in6_addr)];
+
 	    memcpy(tr.icmpSourceAddress, &icmpSourceSocketAddress.sin6_addr, sizeof(in6_addr));
 	    tr.icmpSourceAddressLength = sizeof(in6_addr);
 
-	    tr.icmpType = icmpHeader->icmp6_type;
-	    tr.icmpCode = icmpHeader->icmp6_code;
+	    tr.result = GetAction(icmpHeader->icmp6_type,
+				  icmpHeader->icmp6_code);
 
 	    done = 1;
 
@@ -292,22 +320,13 @@ int PctestIpv6Udp::Test(TestRecord &tr)
 	}
 	else {
 
-	    IF_DEBUG(2, fprintf(stderr, "timeout\n"));
-
-	    tr.targetAddress = new char[sizeof(in6_addr)];
-	    memcpy(tr.targetAddress, &targetAddress, sizeof(in6_addr));
-	    tr.targetAddressLength = sizeof(in6_addr);
-
-	    tr.icmpDestAddress = new char[sizeof(in6_addr)];
-	    memset(tr.icmpDestAddress, 0, sizeof(in6_addr));
-	    tr.icmpDestAddressLength = sizeof(in6_addr);
+	    IF_DEBUG(2, fprintf(stderr, "Timeout\n"));
 
 	    tr.icmpSourceAddress = new char[sizeof(in6_addr)];
 	    memset(tr.icmpSourceAddress, 0, sizeof(in6_addr));
 	    tr.icmpSourceAddressLength = sizeof(in6_addr);
 
-	    tr.icmpType = -1;
-	    tr.icmpCode = -1;
+	    tr.result = PctestActionTimeout;
 
 	    done = 1;
 	}
@@ -318,3 +337,43 @@ int PctestIpv6Udp::Test(TestRecord &tr)
 
 }
 
+//
+// PctestIpv6Udp::GetMinSize
+//
+// Input:  None
+//
+// Output:  Minimum packet size possible for this protocol (in return
+// value).
+//
+int PctestIpv6Udp::GetMinSize() 
+{
+    return (sizeof(ip6_hdr) + sizeof(udphdr) + 4);
+}
+
+//
+// PctestIpv6Udp::GetAction
+//
+// Input:  a test record
+//
+// Output:  action code
+//
+// Figure out the meaning of a particular combination of ICMPv6 type
+// and code values.
+//
+PctestActionType PctestIpv6Udp::GetAction(int icmpType, int icmpCode)
+{
+    if (icmpType == ICMP6_TIME_EXCEEDED) {
+	return PctestActionValid;
+    }
+    else if ((icmpType == ICMP6_DST_UNREACH) &&
+	     (icmpCode == ICMP6_DST_UNREACH_NOPORT)) {
+	return PctestActionValidLasthop;
+    }
+    else if ((icmpType == ICMP6_DST_UNREACH) &&
+	     (icmpCode == ICMP6_DST_UNREACH_ADMIN)) {
+	return PctestActionFiltered;
+    }
+    else {
+	return PctestActionAbort;
+    }
+}
