@@ -1,9 +1,9 @@
-static char rcsid[] = "$Id: main.cc,v 1.54 2000/05/10 20:54:05 bmah Exp $";
+static char rcsid[] = "$Id: main.cc,v 1.82 2001/01/22 17:31:59 bmah Exp $";
 //
-// $Id: main.cc,v 1.54 2000/05/10 20:54:05 bmah Exp $
+// $Id: main.cc,v 1.82 2001/01/22 17:31:59 bmah Exp $
 //
 // main.cc
-// Bruce A. Mah <bmah@ca.sandia.gov>
+// Bruce A. Mah <bmah@acm.org>
 //
 // This work was first produced by an employee of Sandia National
 // Laboratories under a contract with the U.S. Department of Energy.
@@ -49,8 +49,11 @@ extern "C" {
 #include "Pctest.h"
 #include "PctestIpv4Udp.h"
 #include "PctestIpv4Raw.h"
+#include "PctestIpv4Tcp.h"
+#include "PctestIpv4Icmp.h"
 #include "PctestIpv4File.h"
 #ifdef HAVE_IPV6
+#include "PctestIpv6Icmp.h"
 #include "PctestIpv6Udp.h"
 #include "PctestIpv6File.h"
 #endif /* HAVE_IPV6 */
@@ -58,6 +61,14 @@ extern "C" {
 #include "TestRecord.h"
 #ifdef HAVE_SNMP
 #include "GetIfInfo.h"
+#endif /* HAVE_SNMP */
+
+//
+// Forward declarations
+//
+void DoPchar(Pctest *pct);
+void DoTrout(Pctest *pct);
+#ifdef HAVE_SNMP
 void PrintIfInfo(const GetIfInfo *);
 #endif /* HAVE_SNMP */
 
@@ -71,6 +82,8 @@ double Gap = 0.25;
 GapType GapDist = GapFixed;
 unsigned int Hops = 30;
 unsigned int Increment = 32;
+char *OriginHost = NULL;
+ModeType Mode = ModePchar;
 unsigned int Mtu = 1500;
 NetworkProtocolType NetworkProtocol = NetworkProtocolNone;
 int NumericFlag = 0;
@@ -130,8 +143,10 @@ void Usage(char *argv0) {
 #endif /* HAVE_SNMP */
 "[-t timeout] [-T tos] [-v] [-V] [-w file] -r file | host]\n", argv0);
     fprintf(stderr, "\t-a analysis\tSet analysis type (default is lsq)\n");
-    fprintf(stderr, "\t\t\tlsq\tLeast squares linear fit\n");
+    fprintf(stderr, "\t\t\tlsq\tLeast sum of squares linear fit\n");
     fprintf(stderr, "\t\t\tkendall\tLinear fit using Kendall's test statistic\n");
+    fprintf(stderr, "\t\t\tlms\tLeast median of squares linear fit\n");
+    fprintf(stderr, "\t\t\tlmsint\tLeast median of squares linear fit (integer computations)\n");
     fprintf(stderr, "\t-c\t\tIgnore route changes (not recommended)\n");
     fprintf(stderr, "\t-d debuglevel\tSet debugging output level\n");
     fprintf(stderr, "\t-g gap\t\tInter-test gap in seconds (default = %0.2f)\n", Gap);
@@ -141,13 +156,20 @@ void Usage(char *argv0) {
     fprintf(stderr, "\t-H hops\t\tMaximum number of hops (default = %d)\n", Hops);
     fprintf(stderr, "\t-h\t\tPrint this help information\n");
     fprintf(stderr, "\t-i increment\tPacket size increment (default = %d)\n", Increment);
+    fprintf(stderr, "\t-l host\t\tSet origin address of probes (defaults to hostname)\n");
     fprintf(stderr, "\t-m mtu\t\tMaximum packet size to check (default = %d)\n", Mtu);
+    fprintf(stderr, "\t-M mode\t\tOperational mode (defaults to pchar)\n");
+    fprintf(stderr, "\t\t\tpchar\tPath characterization\n");
+    fprintf(stderr, "\t\t\ttrout\tTiny traceroute\n");
     fprintf(stderr, "\t-n\t\tDon't resolve addresses to hostnames\n");
     fprintf(stderr, "\t-p protocol\tNetwork protocol (default is ipv4udp)\n");
-    fprintf(stderr, "\t\t\tipv4udp\tUDP over IPv4\n");
-    fprintf(stderr, "\t\t\tipv4raw\traw IPv4 (simulated UDP)\n");
+    fprintf(stderr, "\t\t\tipv4udp\t\tUDP over IPv4\n");
+    fprintf(stderr, "\t\t\tipv4raw\t\tUDP over IPv4 (raw sockets)\n");
+    fprintf(stderr, "\t\t\tipv4tcp\t\tTCP over IPv4 (raw sockets)\n");
+    fprintf(stderr, "\t\t\tipv4icmp\tICMP over IPv4 (raw sockets)\n");
 #ifdef HAVE_IPV6    
-    fprintf(stderr, "\t\t\tipv6udp\tUDP over IPv6\n");
+    fprintf(stderr, "\t\t\tipv6icmp\tICMPv6 over IPv6 (raw sockets)\n");
+    fprintf(stderr, "\t\t\tipv6udp\t\tUDP over IPv6\n");
 #endif /* HAVE_IPV6 */
     fprintf(stderr, "\t-P port\t\tStarting port number (default = %d)\n", Port);
     fprintf(stderr, "\t-q\t\tQuiet output\n");
@@ -236,6 +258,48 @@ char *GetAddressFamily()
 
 }
 
+//
+// GetPrintableNetworkProtocol
+//
+// Input:  Network protocol
+//
+// Output:  ASCII representation of network protocol name.
+//
+char *GetPrintableNetworkProtocol(NetworkProtocolType np)
+{
+    switch (np) {
+    case (NetworkProtocolIpv4Udp): 
+	return("UDP/IPv4");
+	break;
+    case (NetworkProtocolIpv4Raw): 
+	return("UDP/IPv4 (raw sockets)");
+	break;
+    case (NetworkProtocolIpv4Tcp):
+	return("TCP/IPv4 (raw sockets)");
+	break;
+    case (NetworkProtocolIpv4Icmp):
+	return("ICMP/IPv4 (raw sockets)");
+	break;
+    case (NetworkProtocolIpv4File): 
+	return("IPv4 save file");
+	break;
+#ifdef HAVE_IPV6
+    case (NetworkProtocolIpv6Icmp): 
+	return("ICMPv6/IPv6");
+	break;
+    case (NetworkProtocolIpv6Udp): 
+	return("UDP/IPv6");
+	break;
+    case (NetworkProtocolIpv6File): 
+	return("IPv6 save file");
+	break;
+#endif /* HAVE_IPV6 */
+    default:
+	return("unknown network protocol");
+	break;
+    }
+}
+
 //
 // main
 //
@@ -243,12 +307,10 @@ main(int argc, char **argv)
 {
 
     int c;			// getopt
-    int i, j, k;		// universal loop counters
     Pctest *pct = NULL;		// test structure
-    FILE *df = NULL;		// output file 
 
     // Parse command-line arguments using getopt
-    while ((c = getopt(argc, argv, "a:cd:g:G:hH:i:m:np:P:qR:r:s:St:T:vVw:")) != -1) {
+    while ((c = getopt(argc, argv, "a:cd:g:G:hH:i:l:m:M:np:P:qR:r:s:St:T:vVw:")) != -1) {
 
 	// Check for the different command-line flags we accept
 	switch (c) {
@@ -261,11 +323,18 @@ main(int argc, char **argv)
 	    else if (strcasecmp(optarg, "kendall") == 0) {
 		Analysis = AnalysisKendall;
 	    }
+	    else if (strcasecmp(optarg, "lms") == 0) {
+		Analysis = AnalysisLeastMedianSquares;
+	    }
+	    else if (strcasecmp(optarg, "lmsint") == 0) {
+		Analysis = AnalysisLeastMedianSquaresIntegers;
+	    }
 	    else {
 		fprintf(stderr, "Invalid analysis type: %s\n", optarg);
 		Usage(argv[0]);
 		exit(1);
 	    }
+	    break;
 	}
 
 	// c: ignore route changes
@@ -325,12 +394,38 @@ main(int argc, char **argv)
 	    break;
 	}
 
+	// l: set local origin of probes
+	case 'l': {
+	    OriginHost = strdup(optarg);
+	    if (OriginHost == NULL) {
+		fprintf(stderr, "Couldn't allocate space for origin hostname\n");
+		exit(1);
+	    }
+	    break;
+	}
+
 	// m: MTU size
 	case 'm': {
 	    Mtu = atoi(optarg);
 	    break;
 	}
 
+	case 'M': {
+	    if ((strcasecmp(optarg, "pchar") == 0) ||
+	        (strcasecmp(optarg, "pathchar") == 0)) {
+		Mode = ModePchar;
+	    }
+	    else if (strcasecmp(optarg, "trout") == 0) {
+		Mode = ModeTrout;
+	    }
+	    else {
+		fprintf(stderr, "Invalid operational mode %s\n", optarg);
+		Usage(argv[0]);
+		exit(1);
+	    }
+	    break;
+	}
+
 	// n: don't resolve addresses to hostnames
 	case 'n': {
 	    NumericFlag = 1;
@@ -345,7 +440,16 @@ main(int argc, char **argv)
 	    else if (strcasecmp(optarg, "ipv4raw") == 0) {
 		NetworkProtocol = NetworkProtocolIpv4Raw;
 	    }
+	    else if (strcasecmp(optarg, "ipv4tcp") == 0) {
+		NetworkProtocol = NetworkProtocolIpv4Tcp;
+	    }
+	    else if (strcasecmp(optarg, "ipv4icmp") == 0) {
+		NetworkProtocol = NetworkProtocolIpv4Icmp;
+	    }
 #ifdef HAVE_IPV6
+	    else if (strcasecmp(optarg, "ipv6icmp") == 0) {
+		NetworkProtocol = NetworkProtocolIpv6Icmp;
+	    }
 	    else if (strcasecmp(optarg, "ipv6udp") == 0) {
 		NetworkProtocol = NetworkProtocolIpv6Udp;
 	    }
@@ -486,6 +590,10 @@ main(int argc, char **argv)
     TargetHost = argv[optind];
     IF_DEBUG(1, fprintf(stderr, "TargetHost %s\n", TargetHost));
 
+    // Other initialization before we go off to deal with protocol-
+    // dependent things:  Weakly seed process random number generator.
+    srandom(time(NULL));
+
     // Initialize a test structure.  Note that reading from a file
     // is a special case.
     if (ReadFilename) {
@@ -596,20 +704,79 @@ main(int argc, char **argv)
 	case (NetworkProtocolIpv4Raw):
 	    pct = new PctestIpv4Raw(Port);
 	    break;
+	case (NetworkProtocolIpv4Tcp):
+	    pct = new PctestIpv4Tcp(Port);
+	    break;
+	case (NetworkProtocolIpv4Icmp):
+	    pct = new PctestIpv4Icmp();
+	    break;
 #ifdef HAVE_IPV6
+	case (NetworkProtocolIpv6Icmp):
+	    pct = new PctestIpv6Icmp();
+	    break;
 	case (NetworkProtocolIpv6Udp):
 	    pct = new PctestIpv6Udp(Port);
 	    break;
 #endif /* HAVE_IPV6 */
 	default:
 	    fprintf(stderr, "Unknown protocol type...exiting...\n");
-	    break;
+	    exit(1);
 	}
     }
 
+    // Having figured out what the address family is, do the necessary
+    // name resolution to determine the source and destination of
+    // our probe packets.
     if (pct->SetTargetName(TargetHost) < 0) {
 	exit(1);
     }
+    if (pct->SetOriginName(OriginHost) < 0) {
+	exit(1);
+    }
+
+    // Get sockets
+    if (pct->GetSocketOut() < 0) {
+	exit(1);
+    }
+    if (pct->GetSocketIn() < 0) {
+	exit(1);
+    }
+
+    // With all arguments parsed, determine which set of tests
+    // to run and go do it.
+    switch (Mode) {
+
+    case (ModePchar):
+	DoPchar(pct);
+	break;
+
+    case (ModeTrout):
+	DoTrout(pct);
+	break;
+
+    default:
+	fprintf(stderr, "Unknown mode type...exiting...\n");
+	exit(1);
+    }
+
+    // Stick a fork in us, we're done.
+    exit(0);
+}
+
+//
+// DoPchar
+//
+// Input:  Pctest structure controlling the type of tests to run.
+//
+// Output:  None.
+//
+// Run the original path characterization measurement and analysis
+// algorithm.
+//
+void DoPchar(Pctest *pct) 
+{
+    int i, j, k;		// universal loop counters
+    FILE *df = NULL;		// output file 
 
     // Generate set of packet sizes to test.  We'll test packets
     // from Increment to the maximum multiple of Increment that
@@ -649,31 +816,11 @@ main(int argc, char **argv)
     // Start output
     //
     if (!QuietFlag) {
-	fprintf(stdout, "pchar to %s (%s) using ", pct->GetTargetName(), 
-		pct->GetPrintableAddress());
-	switch (NetworkProtocol) {
-	case (NetworkProtocolIpv4Udp): 
-	    fprintf(stdout, "UDP/IPv4\n");
-	    break;
-	case (NetworkProtocolIpv4Raw): 
-	    fprintf(stdout, "raw IP (simulated UDP)\n");
-	    break;
-	case (NetworkProtocolIpv4File): 
-	    fprintf(stdout, "IPv4 save file\n");
-	    break;
-#ifdef HAVE_IPV6
-	case (NetworkProtocolIpv6Udp): 
-	    fprintf(stdout, "UDP/IPv6\n");
-	    break;
-	case (NetworkProtocolIpv6File): 
-	    fprintf(stdout, "IPv6 save file\n");
-	    break;
-#endif /* HAVE_IPV6 */
-	default:
-	    fprintf(stdout, "unknown network protocol\n");
-	    break;
-	}
-	fprintf(stdout, "Packet size increments by %d to %d\n", Increment, Mtu);
+	fprintf(stdout, "pchar to %s (%s) using %s\n", pct->GetTargetName(), 
+		pct->GetPrintableAddress(),
+		GetPrintableNetworkProtocol(NetworkProtocol));
+	fprintf(stdout, "Packet size increments from %d to %d by %d\n", 
+		pct->GetMinSize(), Mtu, Increment);
 	fprintf(stdout, "%d test(s) per repetition\n", testsPerRep);
 	fprintf(stdout, "%d repetition(s) per hop\n", Repetitions);
     }
@@ -706,7 +853,7 @@ main(int argc, char **argv)
 		exit(1);
 	    }
 	    
-	    if (tr->icmpType < 0) {
+	    if (tr->result == PctestActionTimeout) {
 		timeouts++;
 	    }
 	    else {
@@ -714,47 +861,42 @@ main(int argc, char **argv)
 	    }
 	}
 
-	// Print first-hop output line.  Note that we now send off some
-	// packet(s) to ping the destination host, we might be able to
-	// read the address of our outgoing interface by looking at the
-	// destination of the ICMP ping packets.
-	if (timeouts < maxtimeouts) {
-
-	    if (!QuietFlag) {
-		fprintf(stdout, "%2d: %s ", StartHop - 1, 
-			pct->GetPrintableAddress(tr->icmpDestAddress));
-		if (NumericFlag) {
-		    fprintf(stdout, "(%s)", 
-			    pct->GetPrintableAddress(tr->icmpDestAddress));
-		}
-		else {
-		    fprintf(stdout, "(%s)", pct->GetName(tr->icmpDestAddress));
-		}
-		fprintf(stdout, "\n");
-#ifdef HAVE_SNMP
-		if (!ReadFilename && SnmpFlag) {
-		    GetIfInfo *gifp;
-		    gifp = new GetIfInfo(tr->icmpDestAddress, pct);
-		    PrintIfInfo(gifp);
-		    delete gifp;
-		}
-#endif /* HAVE_SNMP */
-
-	    }
-
-	}
-	else {
-	    fprintf(stdout, "Warning: target host did not respond to initial test.\n");
-	    fprintf(stdout, "0: \n");
+        // If we didn't get a response to
+	// our initial pings, then warn about this fact.  
+	if (timeouts >= maxtimeouts) {
+	    fprintf(stdout, 
+		    "Warning: target host did not respond to initial test.\n");
 	}
 
 	delete tr;
 
     }
-    else {
-	fprintf(stdout, "0: \n");
+
+    // Print first-hop output line.  Prior versions
+    // of this code were more complex because they used the returned
+    // ICMP packets to figure out the source address.  But now we
+    // have that information up-front.
+    if (!QuietFlag) {
+	fprintf(stdout, "%2d: %s ", StartHop - 1, 
+		pct->GetPrintableAddress(pct->GetOriginAddress()));
+	if (NumericFlag) {
+	    fprintf(stdout, "(%s)", 
+		    pct->GetPrintableAddress(pct->GetOriginAddress()));
+	}
+	else {
+	    fprintf(stdout, "(%s)", pct->GetOriginName());
+	}
+	fprintf(stdout, "\n");
+#ifdef HAVE_SNMP
+	if (!ReadFilename && SnmpFlag) {
+	    GetIfInfo *gifp;
+	    gifp = new GetIfInfo(pct->GetOriginAddress(), pct);
+	    PrintIfInfo(gifp);
+	    delete gifp;
+	}
+#endif /* HAVE_SNMP */
     }
-    
+
     // Initialize statistics save file and write some stuff out
     if (WriteFilename) {
 
@@ -770,9 +912,12 @@ main(int argc, char **argv)
 	}
 
 	// Write initial parameters to the file
-	fprintf(df, "addresses %s\n", pct->GetAddressFamily());
+	fprintf(df, "addresses %s\n", pct->GetAddressFamilyString());
 	fprintf(df, "targethost %s\n", pct->GetTargetName());
+	fprintf(df, "src %s\n", pct->GetPrintableAddress(pct->GetOriginAddress()));
+	fprintf(df, "dest %s\n", pct->GetPrintableAddress());
 	fprintf(df, "hops %d\n", Hops);
+	fprintf(df, "minsize %d\n", pct->GetMinSize());
 	fprintf(df, "increment %d\n", Increment);
 	fprintf(df, "mtu %d\n", Mtu);
 	fprintf(df, "repetitions %d\n", Repetitions);
@@ -785,7 +930,7 @@ main(int argc, char **argv)
     // keep the hop number minus StartHop, and add it back before printing,
     // since the internal data structures we use start from zero.)
     //
-    int lastHopFlag;		// *could* be the last hop
+    bool lastHopFlag;		// *could* be the last hop
     double aCumulativeLast = 0.0,
 	   bCumulativeLast = 0.0,
 	   r2CumulativeLast = 0.0;
@@ -808,7 +953,7 @@ main(int argc, char **argv)
 
 	// Initialize statistics structure
 	PartialResults[i] = new ResultTable(Increment, Mtu, Repetitions);
-	lastHopFlag = 1;
+	lastHopFlag = true;
 
 	// Run the correct number of repetitions and iterate over the
 	// packet sizes.
@@ -850,7 +995,7 @@ main(int argc, char **argv)
 
 		// Check for a timeout.  If we got one, then increment
 		// the appropriate counter and keep going.
-		if (tr->icmpType < 0) {
+		if (tr->result == PctestActionTimeout) {
 		    packetsLost++;
 		    delete tr;
 		    continue;
@@ -870,7 +1015,7 @@ main(int argc, char **argv)
 		    realTimestamp = true;
 		}
 
-		IF_DEBUG(2, fprintf(stderr, "bytes = %d, rtt = %f, ip_src = %s\n", tr->size, rtt, pct->GetPrintableAddress(tr->icmpSourceAddress)));
+		IF_DEBUG(2, fprintf(stderr, "bytes = %d, rtt = %f, ip_src = %s, replbytes = %d\n", tr->size, rtt, pct->GetPrintableAddress(tr->icmpSourceAddress), tr->replsize));
 
 		// If address changed, we're in trouble because the route
 		// changed.
@@ -895,14 +1040,15 @@ main(int argc, char **argv)
 		// Have the Pctest subclass figure out what the ICMP
 		// type and code fields meant (they're kind of protocol-
 		// independent).
-		pcta = pct->GetAction(tr);
+		pcta = tr->result;
 
 		// If we received a valid (to us) ICMP message, then
 		// attempt to store timing information.
 		if ((pcta == PctestActionValid) || 
 		    (pcta == PctestActionValidLasthop)) {
 
-		    if ((PartialResults[i]->put(tr->size, rtt)) < 0) {
+		    if ((PartialResults[i]->put(tr->size + tr->replsize, 
+						rtt)) < 0) {
 			fprintf(stderr, "Couldn't store result\n");
 			abort();
 		    }
@@ -912,7 +1058,7 @@ main(int argc, char **argv)
 		    // We hit a firewall or something that's going
 		    // to mess with our packets.  Give up now.
 		    fprintf(stderr, "ICMP: packet filtered\n");
-		    lastHopFlag = 1;
+		    lastHopFlag = true;
 		    goto endreps;
 
 		}
@@ -921,15 +1067,14 @@ main(int argc, char **argv)
 		    // in trouble.  It'd be nice if the protocol-dependent
 		    // packet processing threw away everything except
 		    // time exceeded and port-unreachable.
-		    fprintf(stderr, "Unexpected ICMP message %d received\n",
-			    tr->icmpType);
+		    fprintf(stderr, "Unexpected response to probe\n");
 		    abort();
 		}
 
 		// If ICMP type was time exceeded, we know this is *not*
 		// the last hop.
 		if (pcta != PctestActionValidLasthop) {
-		    lastHopFlag = 0;
+		    lastHopFlag = false;
 		}
 
 		// If we're not reading from a file, we need to set
@@ -1012,6 +1157,12 @@ endreps:
 	else if (Analysis == AnalysisKendall) {
 	    PartialMins[i]->tau(aCumulative, bCumulative, bLower, bUpper);
 	}
+	else if (Analysis == AnalysisLeastMedianSquares) {
+	    PartialMins[i]->lms(aCumulative, bCumulative, r2Cumulative);
+	}
+	else if (Analysis == AnalysisLeastMedianSquaresIntegers) {
+	    PartialMins[i]->lmsint(aCumulative, bCumulative, r2Cumulative);
+	}
 	else {
 	    fprintf(stderr, "Unknown statistical analysis type, exiting...\n");
 	    exit(1);
@@ -1074,6 +1225,10 @@ endreps:
 
 		fprintf(stdout, "                       90%% confidence interval is [%f,%f] ms/B\n", bLower * 1000.0, bUpper * 1000.0);
 	    }
+	    else if ((Analysis == AnalysisLeastMedianSquares) ||
+		     (Analysis == AnalysisLeastMedianSquaresIntegers)) {
+		fprintf(stdout, "    Partial char:      rtt = %f ms, (b = %f ms/B), r2 = %f\n", aCumulative*1000.0, bCumulative*1000.0, r2Cumulative);
+	    }
 	    else {
 		fprintf(stderr, "Unknown statistical analysis type, exiting...\n");
 		exit(1);
@@ -1084,7 +1239,22 @@ endreps:
 	    // Hop characterististics don't make any sense for the first
 	    // hop if we start with hop > 1
 	    if ((i > 0) || (StartHop == 1)) {
-		fprintf(stdout, "    Hop char:          rtt = %f ms, bw = %f Kbps\n", aHop*1000.0, hopBandwidth);
+		fprintf(stdout, "    Hop char:          rtt = ");
+		if (aHop >= 0.0) {
+		    fprintf(stdout, "%f", aHop*1000.0);
+		}
+		else {
+		    fprintf(stdout, "--.---");
+		}
+		fprintf(stdout, " ms, bw = ");
+		if (hopBandwidth >= 0.0) {
+		    fprintf(stdout, "%f", hopBandwidth);
+		}
+		else {
+		    fprintf(stdout, "--.---");
+		}
+		fprintf(stdout, " Kbps\n");
+
 		fprintf(stdout, "    Hop queueing:      avg = %f ms (%d bytes)\n", qTimeHop, qBytes);
 
 	    }
@@ -1133,12 +1303,43 @@ endreps:
 
 	    double aPath, bPath, r2Path, qPath;
 	    double saPath, sbPath;
-	    PartialMins[i]->slr(aPath, bPath, r2Path, saPath, sbPath);
+	    double bLowerPath, bUpperPath;
+
+	    if (Analysis == AnalysisLeastSquares) {
+		PartialMins[i]->slr(aPath, bPath, r2Path, saPath, sbPath);
+	    }
+	    else if (Analysis == AnalysisKendall) {
+		PartialMins[i]->tau(aPath, bPath, bLowerPath, bUpperPath);
+	    }
+	    else if (Analysis == AnalysisLeastMedianSquares) {
+		PartialMins[i]->lms(aPath, bPath, r2Path);
+	    }
+	    else if (Analysis == AnalysisLeastMedianSquaresIntegers) {
+		PartialMins[i]->lmsint(aPath, bPath, r2Path);
+	    }
+	    else {
+		fprintf(stderr, "Unknown statistical analysis type, exiting...\n");
+		exit(1);
+	    }
+
 	    qPath = PartialResults[i]->queueing();
 	    
 	    fprintf(stdout, "    Path length:       %d hops\n", pathLength);
 	
-	    fprintf(stdout, "    Path char:         rtt = %f ms, r2 = %f\n", aPath*1000.0, r2Path);
+	    if (Analysis == AnalysisLeastSquares) {
+		fprintf(stdout, "    Path char:         rtt = %f ms r2 = %f\n", aPath*1000.0, r2Path);
+	    }
+	    else if (Analysis == AnalysisKendall) {
+		fprintf(stdout, "    Path char:         rtt = %f ms\n", aPath*1000.0);
+	    }
+	    else if ((Analysis == AnalysisLeastMedianSquares) ||
+		     (Analysis == AnalysisLeastMedianSquaresIntegers)) {
+		fprintf(stdout, "    Path char:         rtt = %f ms, r2 = %f\n", aPath*1000.0, r2Path);
+	    }
+	    else {
+		fprintf(stderr, "Unknown statistical analysis type, exiting...\n");
+		exit(1);
+	    }
 	    
 	    if (minBandwidth > 0.0) {
 		fprintf(stdout, "    Path bottleneck:   %f Kbps\n", minBandwidth);
@@ -1155,12 +1356,281 @@ endreps:
     if (WriteFilename) {
 	fclose(df);
     }
-    
-    exit(0);
+}
 
+//
+// DoTrout
+//
+// Input:  Pctest structure controlling the type of tests to run.
+//
+// Output:  None.
+//
+// Run a tiny traceroute (intended for use as a part of a larger
+// measurement infrastructure).
+//
+void DoTrout(Pctest *pct)
+{
+    int i;			// universal loop counters
+    FILE *df = NULL;		// output file 
+
+    //
+    // Some of the original pchar parameters may need adjusting...
+    Repetitions = 1;		// force repetitions to 1
+
+    //
+    // Start output
+    //
+    if (!QuietFlag) {
+	fprintf(stdout, "trout to %s (%s) using %s\n", pct->GetTargetName(), 
+		pct->GetPrintableAddress(), 
+		GetPrintableNetworkProtocol(NetworkProtocol));
+	fprintf(stdout, "Packet size increments from %d to %d by %d\n", 
+		pct->GetMinSize(), Mtu, Increment);
+    }
+
+    // Print first-hop output line.
+    if (!QuietFlag) {
+	fprintf(stdout, "%2d: %s ", StartHop - 1, 
+		pct->GetPrintableAddress(pct->GetOriginAddress()));
+	if (NumericFlag) {
+	    fprintf(stdout, "(%s)", 
+		    pct->GetPrintableAddress(pct->GetOriginAddress()));
+	}
+	else {
+	    fprintf(stdout, "(%s)", pct->GetOriginName());
+	}
+	fprintf(stdout, "\n");
+#ifdef HAVE_SNMP
+	if (!ReadFilename && SnmpFlag) {
+	    GetIfInfo *gifp;
+	    gifp = new GetIfInfo(pct->GetOriginAddress(), pct);
+	    PrintIfInfo(gifp);
+	    delete gifp;
+	}
+#endif /* HAVE_SNMP */
+    }
+
+    // Initialize statistics save file and write some stuff out
+    if (WriteFilename) {
+
+	if (strcmp(WriteFilename, "-") == 0) {
+	    df = stdout;
+	}
+	else {
+	    df = fopen(WriteFilename, "w");
+	    if (df == NULL) {
+		perror("fopen");
+		exit(1);
+	    }
+	}
+
+	// Write initial parameters to the file
+	fprintf(df, "addresses %s\n", pct->GetAddressFamilyString());
+	fprintf(df, "targethost %s\n", pct->GetTargetName());
+	fprintf(df, "src %s\n", pct->GetPrintableAddress(pct->GetOriginAddress()));
+	fprintf(df, "dest %s\n", pct->GetPrintableAddress());
+	fprintf(df, "hops %d\n", Hops);
+	fprintf(df, "minsize %d\n", pct->GetMinSize());
+	fprintf(df, "increment %d\n", Increment);
+	fprintf(df, "mtu %d\n", Mtu);
+	fprintf(df, "repetitions %d\n", Repetitions);
+	fprintf(df, "starthop %d\n", StartHop);
+    }
+
+    //
+    // Begin testing.  We increment the hop number, iterating over
+    // the different packet sizes we have available.  (Really, we
+    // keep the hop number minus StartHop, and add it back before printing,
+    // since the internal data structures we use start from zero.)
+    //
+    bool lastHopFlag;		// *could* be the last hop
+    int packetsLost = 0;
+    int packetsSent = 0;
+
+    for (i = 0; i < Hops; i++) {
+
+	unsigned int packetSize;
+	PctestActionType pcta;
+
+	lastHopFlag = true;
+
+	int rc;		// syscall return code
+	double rtt;
+	TestRecord *tr;	// test record for this test
+
+	// Generate packet size
+	packetSize = Increment * (random() % (Mtu/Increment));
+	if (packetSize < pct->GetMinSize()) {
+	    packetSize = pct->GetMinSize();
+	}
+
+	// Let the user know what we're doing
+	if (VerboseFlag) {
+	    fprintf(stderr, "    hop %2d  size %5d\r", 
+		    i+StartHop, packetSize);
+	}
+
+	// Set up variables for a test
+	tr = new TestRecord;
+	tr->size = packetSize;
+	tr->hops = i+StartHop;
+	gettimeofday(&tr->tvstart, NULL);
+	
+	rc = pct->Test(*tr);
+	packetsSent++;
+	if (rc < 0) {
+	    exit(1);
+	}
+
+	// Write to disk if requested.
+	if (WriteFilename) {
+	    fprintf(df, "%s\n", tr->htoa(pct));
+	}
+
+	// Check for a timeout.  If we got one, then increment
+	// the appropriate counter and keep going.
+	if (tr->result == PctestActionTimeout) {
+	    packetsLost++;
+	    delete tr;
+	    continue;
+	    // XXX Do we want to retry this probe?
+	}
+
+	// Process the results we got back.
+	rtt = tr->tv.tv_sec + (tr->tv.tv_usec/1000000.0);
+
+	IF_DEBUG(2, fprintf(stderr, "bytes = %d, rtt = %f, ip_src = %s, replbytes = %d\n", tr->size, rtt, pct->GetPrintableAddress(tr->icmpSourceAddress), tr->replsize));
+	
+	// Have the Pctest subclass figure out what the ICMP
+	// type and code fields meant (they're kind of protocol-
+	// independent).
+	pcta = tr->result;
+
+	// If we received a valid (to us) ICMP message, then
+	// attempt to store timing information.
+	if ((pcta == PctestActionValid) || 
+	    (pcta == PctestActionValidLasthop)) {
+	  // XXX Do we write out something here?
+	}
+	else if (pcta == PctestActionFiltered) {
+
+	    // We hit a firewall or something that's going
+	    // to mess with our packets.  Give up now.
+	    fprintf(stderr, "ICMP: packet filtered\n");
+	    lastHopFlag = true;
+	    goto endreps;
+	    
+	}
+	else {
+	    // With certain types/codes returned by ICMP, we're also
+	    // in trouble.  It'd be nice if the protocol-dependent
+	    // packet processing threw away everything except
+	    // time exceeded and port-unreachable.
+	    fprintf(stderr, "Unexpected response to probe\n");
+	    abort();
+	}
+
+	// If ICMP type was time exceeded, we know this is *not*
+	// the last hop.
+	if (pcta != PctestActionValidLasthop) {
+	    lastHopFlag = false;
+	}
+
+	// Delay between packets
+	struct timeval tvGap;
+	    
+	if (GapDist == GapFixed) {
+	    
+	    tvGap.tv_sec = (long) Gap;
+	    tvGap.tv_usec = (long) ((Gap - ((long) Gap)) * 1000000.0);
+	    
+	}
+	else if (GapDist == GapExponential) {
+	    
+	    // Generate exponentially distributed random
+	    // value with a mean of Gap.
+	    long u;
+	    double udub;
+	    const double maxudub = pow(2.0, 31.0); // 2^31
+	    double x;
+	    
+	    u = random();
+	    udub = ((double) u / maxudub);
+	    x = - Gap * log(udub);
+	    
+	    IF_DEBUG(2, fprintf(stderr, "main:  exponential gap %f\n", x));
+	    
+	    tvGap.tv_sec = (long) x;
+	    tvGap.tv_usec = (long) ((x - ((long) x)) * 1000000.0);
+	    
+	}
+	else {
+	    
+	    fprintf(stderr, "Unknown gap type\n");
+	    exit(1);
+	    
+	}
+	
+	// Delay for some amount of time.
+	rc = select(0, NULL, NULL, NULL, &tvGap);
+	if (rc < 0) {
+	    perror("select");
+	}
+	
+endreps:
+	// If we weren't in quiet mode, then clear the line we were on
+	if (VerboseFlag) {
+	    fprintf(stdout, "%80s\r", "");
+	}
+	
+	// Per-hop output
+	if (!QuietFlag) {
+	    fprintf(stdout, "%2d: %s ", i + StartHop, 
+		    pct->GetPrintableAddress(tr->icmpSourceAddress));
+	    if (NumericFlag) {
+		fprintf(stdout, "(%s)", 
+			pct->GetPrintableAddress(tr->icmpSourceAddress));
+	    }
+	    else {
+		fprintf(stdout, "(%s)", pct->GetName(tr->icmpSourceAddress));
+	    }
+	    fprintf(stdout, " %d -> %d bytes: %0.3f ms\n", 
+		    tr->size, tr->replsize, rtt * 1000.0);
+	}
+	
+#ifdef HAVE_SNMP
+	if (!ReadFilename && SnmpFlag) {
+	    GetIfInfo *gifp;
+	    gifp = new GetIfInfo(checkAddress, pct);
+	    PrintIfInfo(gifp);
+	    delete gifp;
+	}
+#endif /* HAVE_SNMP */
+	
+	delete tr;
+
+	// If all ICMP messages received were "port unreachable", then
+	// we're done.
+	if (lastHopFlag) {
+	    break;
+	}
+    }
+    
+    int pathLength = i + StartHop;
+
+    // End-of-run processing
+    if (!QuietFlag) {
+    }
+
+    if (WriteFilename) {
+	fclose(df);
+    }
 }
 
 #ifdef HAVE_SNMP
+//
+// PrintIfInfo
+//
 void PrintIfInfo(const GetIfInfo *gifp)
 {
     if (gifp == NULL) {
