@@ -1,8 +1,8 @@
 // -*- c++ -*-
-// $Id: pc.h,v 1.19 2000/05/05 22:26:50 bmah Exp $
+// $Id: pc.h,v 1.25 2001/01/10 16:08:50 bmah Exp $
 //
 // pc.h
-// Bruce A. Mah <bmah@ca.sandia.gov>
+// Bruce A. Mah <bmah@acm.org>
 //
 // This work was first produced by an employee of Sandia National
 // Laboratories under a contract with the U.S. Department of Energy.
@@ -28,16 +28,25 @@ extern int DebugLevel;
 typedef enum {false, true} bool;
 #endif /* SIZEOF_BOOL */
 
+// Mode types
+typedef enum {
+    ModeNone,
+    ModePchar,
+    ModeTrout
+} ModeType;
+
 // Network protocol types
 typedef enum {
     NetworkProtocolNone,
     NetworkProtocolIpv4Udp,
     NetworkProtocolIpv4Raw,
+    NetworkProtocolIpv4Tcp,
+    NetworkProtocolIpv4Icmp,
     NetworkProtocolIpv4File
 #ifdef HAVE_IPV6
     ,
+    NetworkProtocolIpv6Icmp,
     NetworkProtocolIpv6Udp,
-    NetworkProtocolIpv6Raw,
     NetworkProtocolIpv6File
 #endif /* HAVE_IPV6 */
 } NetworkProtocolType;
@@ -46,7 +55,9 @@ typedef enum {
 typedef enum {
     AnalysisNone,
     AnalysisLeastSquares,
-    AnalysisKendall
+    AnalysisKendall,
+    AnalysisLeastMedianSquares,
+    AnalysisLeastMedianSquaresIntegers
 } AnalysisType;
 
 // Gap types
